# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Tests for `caso.extract.manager` module.
"""

import datetime
import uuid

import dateutil.parser
import mock

from caso.extract import manager
from caso.tests import base


class TestCasoManager(base.TestCase):
    def setUp(self):
        super(TestCasoManager, self).setUp()
        manager.SUPPORTED_EXTRACTORS = {"nova": "foo.Bar"}
        self.flags(extractor="nova")
        self.p_extractor = mock.patch.dict('sys.modules',
                                           {'foo': mock.MagicMock()})
        self.p_extractor.start()
        self.manager = manager.Manager()

    def tearDown(self):
        self.p_extractor.stop()
        self.reset_flags()

        super(TestCasoManager, self).tearDown()

    def test_extract_empty_projects(self):
        self.flags(projects=[])

        with mock.patch.object(self.manager.extractor,
                               "extract_for_project") as m:
            self.manager._extract("1999-12-19", "2015-10-23")
            self.assertFalse(m.called)
        self.assertEqual({}, self.manager.records)

    def test_extract(self):
        records = {uuid.uuid4().hex: None}
        self.flags(projects=["bazonk"])

        with mock.patch.object(self.manager.extractor,
                               "extract_for_project") as m:
            m.return_value = records
            self.manager._extract("1999-12-19", "2015-12-19")
            m.assert_called_once_with("bazonk", "1999-12-19", "2015-12-19")
        self.assertEqual(records, self.manager.records)

    def test_get_records_wrong_extract_from(self):
        self.flags(extract_from="1999-12-99")
        self.assertRaises(ValueError,
                          self.manager.get_records)

    def test_get_records_wrong_extract_to(self):
        self.flags(extract_to="1999-12-99")
        self.assertRaises(ValueError,
                          self.manager.get_records)

    def test_get_records_wrong_lastrun(self):
        self.assertRaises(ValueError,
                          self.manager.get_records,
                          lastrun="1999-12-99")

    def test_get_records_with_extract_from_and_to(self):
        date = "1999-12-11"
        end_date = "2015-12-11"
        dt = dateutil.parser.parse(date)
        parsed_end_date = dateutil.parser.parse(end_date)
        self.flags(extract_from=date, extract_to=end_date)

        with mock.patch.object(self.manager, "_extract") as m:
            self.manager.get_records()
            m.assert_called_with(dt, parsed_end_date)

    def test_get_records_with_lastrun(self):
        date = "1999-12-11"

        dt = dateutil.parser.parse(date)
        mock_now = datetime.datetime(2016, 4, 19, 15, 4, 14, 468060)

        with mock.patch.object(self.manager, "_extract") as m:
            with mock.patch('caso.extract.manager.datetime') as MockDatetime:
                MockDatetime.datetime.utcnow.return_value = mock_now

                self.manager.get_records(lastrun=date)
                m.assert_called_with(dt, mock_now)



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

import caso.extract.manager
import caso.keystone_client
import caso.manager
import caso.messenger.logstash
import caso.messenger.ssm


def list_opts():
    return [
        ('DEFAULT', itertools.chain(caso.manager.opts,
                                    caso.manager.cli_opts,
                                    caso.extract.manager.opts,
                                    caso.extract.manager.cli_opts)
         ),
        ('keystone_auth', caso.keystone_client.opts),
        ('logstash', caso.messenger.logstash.opts),
        ('ssm', caso.messenger.ssm.opts),
    ]



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import json
import warnings

from oslo_config import cfg
from oslo_log import log
import six

from caso import keystone_client

CONF = cfg.CONF
CONF.import_opt("mapping_file", "caso.extract.manager")

LOG = log.getLogger(__name__)

openstack_vm_statuses = {
    'active': 'started',
    'build': 'started',
    'confirming_resize': 'started',
    'deleted': 'completed',
    'error': 'error',
    'hard_reboot': 'started',
    'migrating': 'started',
    'password': 'started',
    'paused': 'paused',
    'reboot': 'started',
    'rebuild': 'started',
    'rescue': 'started',
    'resize': 'started',
    'revert_resize': 'started',
    'verify_resize': 'started',
    'shutoff': 'completed',
    'suspended': 'suspended',
    'terminated': 'completed',
    'stopped': 'stopped',
    'saving': 'started',
    'unknown': 'unknown',
}


@six.add_metaclass(abc.ABCMeta)
class BaseExtractor(object):
    def __init__(self):
        try:
            mapping = json.loads(open(CONF.mapping_file).read())
        except ValueError:
            # FIXME(aloga): raise a proper exception here
            raise
        else:
            self.voms_map = {}
            for vo, vomap in six.iteritems(mapping):
                tenant = vomap.get("tenant", None)
                tenants = vomap.get("tenants", [])
                if tenant is not None:
                    warnings.warn("Using deprecated 'tenant' mapping, please "
                                  "use 'projects' instead", DeprecationWarning)
                if tenants:
                    warnings.warn("Using deprecated 'tenants' mapping, please "
                                  "use 'projects' instead", DeprecationWarning)
                tenants.append(tenant)
                projects = vomap.get("projects", tenants)
                if not projects:
                    LOG.warning("No project mapping found for VO %s" % vo)
                for project in projects:
                    self.voms_map[project] = vo

    def _get_keystone_client(self, project):
        client = keystone_client.get_client(CONF, project)
        return client

    def _get_keystone_users(self, ks_client):
        project_id = ks_client.project_id
        users = ks_client.users.list(project_id=project_id)
        return {u.id: u.name for u in users}

    def vm_status(self, status):
        """Return the status corresponding to the OpenStack status.

        :param status: OpenStack status.
        """
        return openstack_vm_statuses.get(status.lower(), 'unknown')

    @abc.abstractmethod
    def extract_for_project(self, project, extract_from):
        """Extract records for a project from given date.

        :param project: Project to extract records for.
        :param extract_from: datetime.datetime object indicating the date to
                             extract records from
        :returns: A dictionary of {"server_id": caso.record.Record"}

        This method should be overriden in a subclass.
        """



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Tests for `caso.record` module.
"""

import uuid

import caso
from caso import exception
from caso import record
from caso.tests import base


class TestCasoManager(base.TestCase):
    def test_invalid_version(self):
        r = record.CloudRecord(uuid.uuid4().hex,
                               "site-foo",
                               "name-foo",
                               uuid.uuid4().hex,
                               uuid.uuid4().hex,
                               "/Foo/User/Fqan")

        self.assertRaises(exception.RecordVersionNotFound,
                          r.as_dict, version="0.0")

        self.assertRaises(exception.RecordVersionNotFound,
                          r.as_json, version="0.0")

    def test_required_fields(self):
        server_id = uuid.uuid4().hex
        site_name = "site-foo"
        server_name = "name-foo"
        server_user_id = uuid.uuid4().hex
        server_project_id = uuid.uuid4().hex
        fqan = "FooVO"
        status = 'completed'
        image_id = uuid.uuid4().hex
        user_dn = "/Foo/bar/baz"

        expected = {
            'FQAN': fqan,
            'GlobalUserName': user_dn,
            'ImageId': image_id,
            'LocalGroupId': server_project_id,
            'LocalUserId': server_user_id,
            'MachineName': server_name,
            'SiteName': site_name,
            'Status': status,
            'VMUUID': server_id,
            'CloudType': caso.user_agent,
        }

        expected_02 = {
            'CpuCount': None,
            'CpuDuration': None,
            'Disk': None,
            'EndTime': None,
            'Memory': None,
            'NetworkInbound': None,
            'NetworkOutbound': None,
            'NetworkType': None,
            'StartTime': None,
            'StorageRecordId': None,
            'SuspendDuration': None,
            'WallDuration': None,
        }

        expected_04 = {
            'CloudComputeService': None,
            'BenchmarkType': None,
            'Benchmark': None,
            'PublicIPCount': None,
        }

        r = record.CloudRecord(server_id,
                               site_name,
                               server_name,
                               server_user_id,
                               server_project_id,
                               fqan,
                               status=status,
                               image_id=image_id,
                               user_dn=user_dn)

        d_02 = r.as_dict(version="0.2")
        self.assertDictContainsSubset(expected, d_02)
        self.assertDictContainsSubset(expected_02, d_02)

        d_04 = r.as_dict(version="0.4")
        self.assertDictContainsSubset(expected, d_04)
        self.assertDictContainsSubset(expected_04, d_04)



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from oslo_log import log
import six

LOG = log.getLogger(__name__)


class CasoException(Exception):
    msg_fmt = "An unknown exception occurred."

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if not message:
            try:
                message = self.msg_fmt % kwargs
            except Exception:
                exc_info = sys.exc_info()
                # kwargs doesn't match a variable in the message
                # log the issue and the kwargs
                LOG.exception('Exception in string format operation')
                for name, value in six.iteritems(kwargs):
                    LOG.error("%s: %s" % (name, value))
                six.reraise(exc_info[0], exc_info[1], exc_info[2])

        super(CasoException, self).__init__(message)


class RecordVersionNotFound(CasoException):
    msg_fmt = "Version %(version)s of accounting record could not be found."


class ClassNotFound(CasoException):
    msg_fmt = "Class %(class_name)s could not be found: %(exception)s."


class LogstashConnectionError(CasoException):
    msg_fmt = ("Cannot send data to logstash %(host)s:%(port)s, "
               "reason: %(exception)s")



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import operator

import dateutil.parser
import glanceclient.client
import novaclient.client
import novaclient.exceptions
from oslo_config import cfg
from oslo_log import log

from caso.extract import base
from caso import keystone_client
from caso import record

CONF = cfg.CONF
CONF.import_opt("extractor", "caso.extract.manager")
CONF.import_opt("site_name", "caso.extract.manager")
CONF.import_opt("benchmark_name_key", "caso.extract.manager")
CONF.import_opt("benchmark_value_key", "caso.extract.manager")

LOG = log.getLogger(__name__)


class OpenStackExtractor(base.BaseExtractor):
    def __init__(self):
        super(OpenStackExtractor, self).__init__()

    def _get_nova_client(self, project):
        session = keystone_client.get_session(CONF, project)
        return novaclient.client.Client(2, session=session)

    def _get_glance_client(self, project):
        session = keystone_client.get_session(CONF, project)
        return glanceclient.client.Client(2, session=session)

    def build_record(self, server, vo, images, flavors, users):
        server_start = self._get_server_start(server)
        server_end = self._get_server_end(server)

        status = self.vm_status(server.status)

        image_id = None
        if server.image:
            image = images.get(server.image['id'])
            image_id = server.image['id']
            if image:
                if image.get("vmcatcher_event_ad_mpuri", None) is not None:
                    image_id = image.get("vmcatcher_event_ad_mpuri", None)

        flavor = flavors.get(server.flavor["id"])
        if flavor:
            bench_name = flavor["extra"].get(CONF.benchmark_name_key)
            bench_value = flavor["extra"].get(CONF.benchmark_value_key)
            memory = flavor["ram"]
            cpu_count = flavor["vcpus"]
            disk = flavor["disk"] + flavor["OS-FLV-EXT-DATA:ephemeral"]
        else:
            bench_name = bench_value = None
            memory = cpu_count = disk = None

        if not all([bench_name, bench_value]):
            if any([bench_name, bench_value]):
                LOG.warning("Benchmark for flavor %s not properly set" %
                            flavor)
            else:
                LOG.debug("Benchmark information for flavor %s not set,"
                          "plase indicate the corret benchmark_name_key "
                          "and benchmark_value_key in the configuration "
                          "file or set the correct properties in the "
                          "flavor." % flavor)

        r = record.CloudRecord(server.id,
                               CONF.site_name,
                               server.name,
                               server.user_id,
                               server.tenant_id,
                               vo,
                               start_time=server_start,
                               end_time=server_end,
                               compute_service=CONF.service_name,
                               status=status,
                               image_id=image_id,
                               user_dn=users.get(server.user_id, None),
                               benchmark_type=bench_name,
                               benchmark_value=bench_value,
                               memory=memory,
                               cpu_count=cpu_count,
                               disk=disk)
        return r

    @staticmethod
    def _get_server_start(server):
        # We use created, as the start_time may change upon certain actions!
        server_start = dateutil.parser.parse(server.created)
        server_start = server_start.replace(tzinfo=None)
        return server_start

    @staticmethod
    def _get_server_end(server):
        server_end = server.__getattr__('OS-SRV-USG:terminated_at')
        if server_end:
            server_end = dateutil.parser.parse(server_end)
            server_end = server_end.replace(tzinfo=None)
        return server_end

    def extract_for_project(self, project, extract_from, extract_to):
        """Extract records for a project from given date querying nova.

        This method will get information from nova.

        :param project: Project to extract records for.
        :param extract_from: datetime.datetime object indicating the date to
                             extract records from
        :param extract_to: datetime.datetime object indicating the date to
                           extract records to
        :returns: A dictionary of {"server_id": caso.record.Record"}
        """
        # Some API calls do not expect a TZ, so we have to remove the timezone
        # from the dates. We assume that all dates coming from upstream are
        # in UTC TZ.
        extract_from = extract_from.replace(tzinfo=None)

        # Try and except here
        nova = self._get_nova_client(project)
        glance = self._get_glance_client(project)
        ks_client = self._get_keystone_client(project)
        users = self._get_keystone_users(ks_client)
        project_id = nova.client.session.get_project_id()

        flavors = {}
        for flavor in nova.flavors.list():
            flavors[flavor.id] = flavor.to_dict()
            flavors[flavor.id]["extra"] = flavor.get_keys()

        images = {image.id: image for image in glance.images.list()}
        records = {}

        vo = self.voms_map.get(project)

        # We cannot use just 'changes-since' in the servers.list() API query,
        # as it will only include servers that have changed its status after
        # that date. However we cannot just get all the usages and then query
        # server by server, as deleted servers are not returned  by the usages
        # call. Moreover, Nova resets the start_time after performing some
        # actions on the server (rebuild, resize, rescue). If we use that time,
        # we may get a drop in the wall time, as a server that has been resized
        # in the middle of its lifetime will suddenly change its start_time
        #
        # Therefore, what we do is the following (hackish approach)
        #
        # 1.- List all the servers that changed its status after the start time
        #     for the reporting period
        # 2.- Build the records for the period [start, end] using those servers
        # 3.- Get all the usages, being aware that the start time may be wrong
        # 4.- Iter over the usages and:
        # 4.1.- get information for servers that are not returned by the query
        #       in (1), for instance servers that have not changed it status.
        #       We build then the records for those severs
        # 4.2.- For all the servers, adjust the CPU, memory and disk resources
        #       as the flavor may not exist, but we can get those resources
        #       from the usages API.

        # Lets start

        # 1.- List all the deleted servers from that period.
        servers = []
        limit = 200
        marker = None
        # Use a marker and iter over results until we do not have more to get
        while True:
            aux = nova.servers.list(
                search_opts={"changes-since": extract_from},
                limit=limit,
                marker=marker
            )
            servers.extend(aux)

            if len(aux) < limit:
                break
            marker = aux[-1].id

        servers = sorted(servers, key=operator.attrgetter("created"))

        # 2.- Build the records for the period. Drop servers outside the period
        # (we do this manually as we cannot limit the query to a period, only
        # changes after start date).
        for server in servers:

            server_start = self._get_server_start(server)
            server_end = self._get_server_end(server)

            # Some servers may be deleted before 'extract_from' but updated
            # afterwards
            if (server_start > extract_to or
                    (server_end and server_end < extract_from)):
                continue

            records[server.id] = self.build_record(server, vo, images,
                                                   flavors, users)

            # Wall and CPU durations are absolute values, not deltas for the
            # reporting period. The nova API only gives use the usages for the
            # requested period, therefore we need to calculate the wall
            # duration by ourselves, then multiply by the nr of CPUs to get the
            # CPU duration.

            # If the machine has not ended, report consumption until
            # extract_to, otherwise get its consuption by substracting ended -
            # started (done by the record).
            if server_end is None or server_end > extract_to:
                wall = extract_to - server_start
                wall = int(wall.total_seconds())
                records[server.id].wall_duration = wall
                # If we are republishing, the machine reports status completed,
                # but it is not True for this period, so we need to fake the
                # status and remove the end time for the server
                records[server.id].end_time = None

                if records[server.id].status == "completed":
                    records[server.id].status = self.vm_status("active")

        # 3.- Get all the usages for the period
        start = extract_from
        aux = nova.usage.get(project_id, start, extract_to)
        usages = getattr(aux, "server_usages", [])

        # 4.- Iter over the results and
        for usage in usages:
            # 4.1 and 4.2 Get the server if it is not yet there
            if usage["instance_id"] not in records:
                try:
                    server = nova.servers.get(usage["instance_id"])
                except novaclient.exceptions.ClientException as e:
                    LOG.error("Cannot get server '%s' from the Nova API, "
                              "probably because it is an old VM that whose "
                              "metadata is wrong in the DB. There will be no "
                              "record generated for this VM. " %
                              usage["instance_id"])
                    LOG.exception(e)

                    continue

                server_start = self._get_server_start(server)
                if server_start > extract_to:
                    continue
                record = self.build_record(server, vo, images,
                                           flavors, users)

                server_start = record.start_time

                # End time must ben the time when the machine was ended, but it
                # may be none
                if usage.get('ended_at', None) is not None:
                    server_end = dateutil.parser.parse(usage["ended_at"])
                    record.end_time = server_end
                else:
                    server_end = None

                # Wall and CPU durations are absolute values, not deltas for
                # the reporting period. The nova API only gives use the usages
                # for the requested period, therefore we need to calculate the
                # wall duration by ourselves, then multiply by the nr of CPUs
                # to get the CPU duration.

                # If the machine has not ended, report consumption until
                # extract_to, otherwise get its consuption by substracting
                # ended - started (done by the record).
                if server_end is None or server_end > extract_to:
                    wall = extract_to - server_start
                    wall = int(wall.total_seconds())
                    record.wall_duration = wall
                    # If we are republishing, the machine reports status
                    # completed, but it is not True for this period, so we need
                    # to fake the status
                    if record.status == "completed":
                        record.status = self.vm_status("active")

                cput = wall * usage["vcpus"]
                record.cpu_duration = cput

                records[server.id] = record

            # Adjust resources that may not be
            record = records[usage["instance_id"]]
            record.memory = usage["memory_mb"]
            record.cpu_count = usage["vcpus"]
            record.disk = usage["local_gb"]

        return records



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import datetime

import dateutil.parser
from dateutil import tz
from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils
import six

SUPPORTED_EXTRACTORS = {
    'nova': 'caso.extract.nova.OpenStackExtractor',
    'ceilometer': 'caso.extract.ceilometer.CeilometerExtractor',
}

opts = [
    cfg.StrOpt('site_name',
               help='Site name as in GOCDB.'),
    cfg.StrOpt('service_name',
               default='$site_name',
               help='Service name within the site'),
    cfg.ListOpt('projects',
                default=[],
                deprecated_name='tenants',
                help='List of projects to extract accounting records from.'),
    cfg.StrOpt('mapping_file',
               default='/etc/caso/voms.json',
               deprecated_group="extractor",
               help='File containing the VO <-> project mapping as used '
               'in Keystone-VOMS.'),
    cfg.StrOpt('benchmark_name_key',
               default='accounting:benchmark_type',
               help='Metadata key used to retrieve the benchmark type '
                    'from the flavor properties.'),
    cfg.StrOpt('benchmark_value_key',
               default='accounting:benchmark_value',
               help='Metadata key used to retrieve the benchmark value '
                    'from the flavor properties.'),

]

cli_opts = [
    cfg.StrOpt('extract-to',
               deprecated_name='extract_to',
               help='Extract record changes until this date. '
                    'If it is not set, we use now. If a server has '
                    'ended after this date, it will be included, but '
                    'the consuption reported will end on this date. '
                    'If no time zone is specified, UTC will be used.'),
    cfg.StrOpt('extract-from',
               deprecated_name='extract_from',
               help='Extract records that have changed after this date. This '
                    'means that if a record has started before this date, and '
                    'it has changed after this date (i.e. it is still running '
                    'or it has ended) it will be reported. \n'
                    'If it is not set, extract records from last run. '
                    'If it is set to None and last run file is not present, '
                    'it will extract records from the beginning of time. '
                    'If no time zone is specified, UTC will be used.'),
    cfg.StrOpt('extractor',
               choices=SUPPORTED_EXTRACTORS.keys(),
               default='nova',
               help='Which extractor to use for getting the data. '
                    'If you do not specify anything, nova will be '
                    'used.'),
]

CONF = cfg.CONF

CONF.register_opts(opts)
CONF.register_cli_opts(cli_opts)

LOG = log.getLogger(__name__)


class Manager(object):
    def __init__(self):
        extractor_class = importutils.import_class(
            SUPPORTED_EXTRACTORS[CONF.extractor])
        self.extractor = extractor_class()
        self.records = None

    def _extract(self, extract_from, extract_to):
        self.records = {}
        for project in CONF.projects:
            LOG.info("Extracting records for project '%s'" % project)
            try:
                records = self.extractor.extract_for_project(project,
                                                             extract_from,
                                                             extract_to)
            except Exception:
                records = []
                LOG.exception("Cannot extract records for '%s'" % project)
            else:
                LOG.info("Extracted %d records for project '%s' from "
                         "%s to %s" % (len(records), project, extract_from,
                                       extract_to))
            self.records.update(records)

    def get_records(self, lastrun="1970-01-01"):
        """Get records from given date

        :param lastrun: date to get records from (optional).

        If CONF.extract_from is present, it will be used instead of the
        lastrun parameter. If CONF.extract_to is present, it will be used
        instead of the extract_to parameter
        """
        extract_from = CONF.extract_from or lastrun
        extract_to = CONF.extract_to or datetime.datetime.utcnow()

        if isinstance(extract_from, six.string_types):
            extract_from = dateutil.parser.parse(extract_from)
        if isinstance(extract_to, six.string_types):
            extract_to = dateutil.parser.parse(extract_to)

        if extract_from.tzinfo is None:
            extract_from.replace(tzinfo=tz.tzutc())
        if extract_to.tzinfo is None:
            extract_to.replace(tzinfo=tz.tzutc())

        if self.records is None:
            self._extract(extract_from, extract_to)
        return self.records



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import loading
from keystoneclient.v3 import client as ks_client_v3
from oslo_config import cfg

CONF = cfg.CONF

CFG_GROUP = "keystone_auth"

loading.register_auth_conf_options(CONF, CFG_GROUP)
loading.register_session_conf_options(CONF, CFG_GROUP)

opts = (loading.get_auth_common_conf_options() +
        loading.get_session_conf_options() +
        loading.get_auth_plugin_conf_options('password'))


def get_session(conf, project):
    """Get an auth session."""
    auth_plugin = loading.load_auth_from_conf_options(conf, CFG_GROUP,
                                                      project_name=project)
    return loading.load_session_from_conf_options(conf, CFG_GROUP,
                                                  auth=auth_plugin)


def get_client(conf, project):
    """Return a client for Keystone."""
    sess = get_session(conf, project)
    return ks_client_v3.Client(session=sess, interface='public')



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ceilometerclient.client
import dateutil.parser
from oslo_config import cfg
from oslo_log import log

from caso.extract import nova
from caso import keystone_client

CONF = cfg.CONF

LOG = log.getLogger(__name__)


class CeilometerExtractor(nova.OpenStackExtractor):
    def __init__(self):
        super(CeilometerExtractor, self).__init__()

    def _get_ceilometer_client(self, project):
        session = keystone_client.get_session(CONF, project)
        return ceilometerclient.client.get_client('2', session=session)

    def _build_query(self, project_id=None, start_date=None, end_date=None):
        q = []
        if project_id:
            q.append({'field': 'project_id', 'value': project_id})
        if start_date:
            q.append({'field': 'timestamp', 'op': 'ge', 'value': start_date})
        if end_date:
            q.append({'field': 'timestamp', 'op': 'le', 'value': end_date})
        return q

    def _fill_metric(self, metric_name, samples, records,
                     get_id=lambda s: s.resource_id,
                     unit_conv=lambda v: v):
        """Fills a given metric in the records

        get_id is a function that gets the instance id from the sample
        conv is a function that converts the metric to the units
             requested by the usage record

        """

        instance_timestamps = {}
        for sample in samples:
            instance_id = get_id(sample)
            try:
                r = records[instance_id]
            except KeyError:
                continue
            # takes the maximum value from ceilometer
            sample_value = unit_conv(sample.counter_volume)
            instance_ts = instance_timestamps.get(instance_id, None)
            if instance_ts is None:
                r.__dict__[metric_name] = sample_value
            try:
                r.__dict__[metric_name] = max(r.__dict__[metric_name],
                                              sample_value)
            except KeyError:
                r.__dict__[metric_name] = sample_value
            sample_ts = dateutil.parser.parse(sample.timestamp)
            instance_timestamps[instance_id] = sample_ts

    def _fill_cpu_metric(self, cpu_samples, records):
        self._fill_metric('cpu_duration', cpu_samples, records,
                          # convert ns to s
                          unit_conv=lambda v: int(v / 1e9))

    def _fill_net_metric(self, metric_name, net_samples, records):
        self._fill_metric(metric_name, net_samples, records,
                          lambda s: s.resource_metadata.get('instance_id'),
                          # convert bytes to GB
                          unit_conv=lambda v: int(v / 2 ** 30))

    def extract_for_project(self, project, lastrun, extract_to):
        """Extract records for a project from given date.

        This method will get information from nova, and will enhance it with
        information from ceilometer.

        :param project: Project to extract records for.
        :param extract_from: datetime.datetime object indicating the date to
                             extract records from
        :param extract_to: datetime.datetime object indicating the date to
                             extract records to
        :returns: A dictionary of {"server_id": caso.record.Record"}
        """
        records = super(CeilometerExtractor,
                        self).extract_for_project(project, lastrun, extract_to)
        # Try and except here
        ks_conn = self._get_keystone_client(project)
        conn = self._get_ceilometer_client(project)
        # See comment in nova.py, remove TZ from the dates.
        lastrun = lastrun.replace(tzinfo=None)
        search_query = self._build_query(ks_conn.project_id, lastrun,
                                         extract_to)

        cpu = conn.samples.list(meter_name='cpu', q=search_query)
        self._fill_cpu_metric(cpu, records)
        net_in = conn.samples.list(meter_name='network.incoming.bytes',
                                   q=search_query)
        self._fill_net_metric('network_in', net_in, records)
        net_out = conn.samples.list(meter_name='network.outcoming.bytes',
                                    q=search_query)
        self._fill_net_metric('network_out', net_out, records)

        return records



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from caso.extract import nova
from caso.tests import base


class TestCasoManager(base.TestCase):
    def setUp(self):
        super(TestCasoManager, self).setUp()
        self.flags(mapping_file="etc/caso/voms.json.sample")
        self.extractor = nova.OpenStackExtractor()

    def tearDown(self):
        self.reset_flags()

        super(TestCasoManager, self).tearDown()



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import os
import os.path

import dateutil.parser
from dateutil import tz
from oslo_concurrency import lockutils
from oslo_config import cfg

import caso.extract.manager
import caso.messenger
from caso import utils

opts = [
    cfg.ListOpt('messengers',
                default=['caso.messenger.noop.NoopMessenger'],
                help='List of messenger that will dispatch records. '
                'valid values are %s' %
                ["%s.%s" % (i.__module__, i.__name__)
                 for i in caso.messenger.all_managers()]),
    cfg.StrOpt('spooldir',
               default='/var/spool/caso',
               help='Spool directory.'),
]

override_lock = cfg.StrOpt(
    "lock_path",
    default=os.environ.get("CASO_LOCK_PATH", "$spooldir"),
    help="Directory to use for lock files. For security, the specified "
         "directory should only be writable by the user running the "
         "processes that need locking. Defaults to environment variable "
         "CASO_LOCK_PATH or $spooldir"
)
opts.append(override_lock)

cli_opts = [
    cfg.BoolOpt('dry-run',
                deprecated_name='dry_run',
                default=False,
                help='Extract records but do not push records to SSM. This '
                'will not update the last run date.'),
]

CONF = cfg.CONF

CONF.register_opts(opts)
CONF.register_cli_opts(cli_opts)


class Manager(object):
    def __init__(self):
        utils.makedirs(CONF.spooldir)
        self.last_run_file = os.path.join(CONF.spooldir, "lastrun")

        self.extractor_manager = caso.extract.manager.Manager()
        self.messenger = caso.messenger.Manager()

        self.lock_path = CONF.lock_path

    @property
    def lastrun(self):
        if os.path.exists(self.last_run_file):
            with open(self.last_run_file, "r") as fd:
                date = fd.read()
        else:
            date = "1970-01-01"

        try:
            date = dateutil.parser.parse(date)
        except Exception:
            # FIXME(aloga): raise a proper exception here
            raise
        return date

    def run(self):
        @lockutils.synchronized("caso_should_not_run_in_parallel",
                                lock_path=self.lock_path, external=True)
        def synchronized():
            records = self.extractor_manager.get_records(lastrun=self.lastrun)
            if not CONF.dry_run:
                self.messenger.push_to_all(records)
                with open(self.last_run_file, "w") as fd:
                    fd.write(str(datetime.datetime.now(tz.tzutc())))

        return synchronized()



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Tests for `caso.manager` module.
"""

import datetime

from dateutil import tz
import mock
from oslo_concurrency.fixture import lockutils as lock_fixture
import six

from caso import manager
from caso.tests import base


class TestCasoManager(base.TestCase):

    REQUIRES_LOCKING = True

    def setUp(self):
        self.useFixture(lock_fixture.ExternalLockFixture())
        super(TestCasoManager, self).setUp()
        self.patchers = {
            "makedirs": mock.patch('caso.utils.makedirs'),
            "extract": mock.patch('caso.extract.manager.Manager'),
            "messenger": mock.patch('caso.messenger.Manager'),
        }
        self.mocks = {}
        for k, p in six.iteritems(self.patchers):
            self.mocks[k] = p.start()

        self.manager = manager.Manager()

    def tearDown(self):
        for p in self.patchers.values():
            p.stop()

        super(TestCasoManager, self).tearDown()

    def test_lastrun_does_not_exist(self):
        self.manager.last_run_file = "foobarbarz/does/not/exist"
        expected = datetime.datetime(1970, 1, 1, 0, 0)
        self.assertEqual(expected, self.manager.lastrun)

    def test_lastrun_exists(self):
        expected = datetime.datetime(2014, 12, 10, 13, 10, 26, 664598)
        aux = six.StringIO(str(expected))

        if six.PY3:
            builtins_open = 'builtins.open'
        else:
            builtins_open = '__builtin__.open'

        with mock.patch("os.path.exists") as path:
            with mock.patch(builtins_open) as fopen:
                fopen.return_value.__enter__ = lambda x: aux
                fopen.return_value.__exit__ = mock.Mock()
                path.return_value = True

                self.assertEqual(expected, self.manager.lastrun)

    def test_lastrun_is_invalid(self):
        aux = six.StringIO("foo")

        # NOTE(aloga): manager.lastrun is a property, so we need to
        # create our own callable here.
        def call(self):
            return self.manager.lastrun

        if six.PY3:
            builtins_open = 'builtins.open'
        else:
            builtins_open = '__builtin__.open'

        with mock.patch("os.path.exists") as path:
            with mock.patch(builtins_open) as fopen:
                fopen.return_value.__enter__ = lambda x: aux
                fopen.return_value.__exit__ = mock.Mock()
                path.return_value = True

                self.assertRaises(ValueError, call, self)

    def test_dry_run(self):
        self.flags(dry_run=True)
        # NOTE(aloga): cannot patch a property of an instance, see
        # https://code.google.com/p/mock/issues/detail?id=117
        with mock.patch("caso.manager.Manager.lastrun",
                        new_callable=mock.PropertyMock) as lastrun:
            lastrun.return_value = datetime.datetime.now(tz.tzutc())
            mngr = manager.Manager()
            mngr.messenger.push_to_all.assert_not_called()
            mngr.run()
            self.assertFalse(mngr.messenger.push_to_all.called)



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import json
import pprint

import caso
from caso import exception


class CloudRecord(object):
    """The CloudRecord class holds information for each of the records.

    This class is versioned, following the Cloud Accounting Record versions.
    """

    # Version 0.2: initial version
    # Version 0.4: Add 0.4 fields
    version = "0.4"

    _v02_fields = [
        "VMUUID",
        "SiteName",
        "MachineName",
        "LocalUserId",
        "LocalGroupId",
        "GlobalUserName",
        "FQAN",
        "Status",
        "StartTime",
        "EndTime",
        "SuspendDuration",
        "WallDuration",
        "CpuDuration",
        "CpuCount",
        "NetworkType",
        "NetworkInbound",
        "NetworkOutbound",
        "Memory",
        "Disk",
        "StorageRecordId",
        "ImageId",
        "CloudType",
    ]

    _v04_fields = _v02_fields + [
        "CloudComputeService",
        "BenchmarkType",
        "Benchmark",
        "PublicIPCount",
    ]

    _version_field_map = {
        "0.2": _v02_fields,
        "0.4": _v04_fields,
    }

    def __init__(self, uuid, site, name, user_id, group_id, fqan,
                 status=None,
                 start_time=None, end_time=None,
                 suspend_duration=None, wall_duration=None, cpu_duration=None,
                 network_type=None, network_in=None, network_out=None,
                 cpu_count=None, memory=None, disk=None,
                 image_id=None, cloud_type=caso.user_agent,
                 storage_record_id=None,
                 vo=None, vo_group=None, vo_role=None,
                 user_dn=None,
                 compute_service=None,
                 benchmark_value=None, benchmark_type=None,
                 public_ip_count=None):

        self.uuid = uuid
        self.site = site
        self.name = name
        self.user_id = user_id
        self.group_id = group_id
        self.fqan = fqan
        self.status = status
        self.start_time = start_time
        self.end_time = end_time
        self.suspend_duration = suspend_duration
        self.wall_duration = wall_duration
        self.cpu_duration = cpu_duration
        self.network_type = network_type
        self.network_in = network_in
        self.network_out = network_out
        self.cpu_count = cpu_count
        self.memory = memory
        self.disk = disk
        self.image_id = image_id
        self.cloud_type = cloud_type
        self.storage_record_id = storage_record_id
        self.user_dn = user_dn
        self.compute_service = compute_service
        self.benchmark_value = benchmark_value
        self.benchmark_type = benchmark_type
        self.public_ip_count = public_ip_count

    def __repr__(self):
        return pprint.pformat(self.as_dict())

    def as_dict(self, version=None):
        """Return CloudRecord as a dictionary.

        :param str version: optional, if passed it will format the record
                            acording to that account record version

        :returns: A dict containing the record.
        """
        if version is None:
            version = self.version

        if version not in self._version_field_map:
            raise exception.RecordVersionNotFound(version=version)

        return {k: v for k, v in self.map.items()
                if k in self._version_field_map[version]}

    @property
    def wall_duration(self):
        duration = None
        if self._wall_duration is not None:
            duration = self._wall_duration
        elif None not in (self._start_time, self._end_time):
            duration = (self.end_time - self.start_time).total_seconds()
        return int(duration) if duration else duration

    @wall_duration.setter
    def wall_duration(self, value):
        if value and not isinstance(value, (int, float)):
            raise ValueError("Duration must be a number")
        self._wall_duration = value

    @property
    def cpu_duration(self):
        duration = None
        if self._cpu_duration is not None:
            duration = self._cpu_duration
        elif self.wall_duration is not None and self.cpu_count:
            duration = self.wall_duration * self.cpu_count
        return int(duration) if duration else duration

    @cpu_duration.setter
    def cpu_duration(self, value):
        if value and not isinstance(value, (int, float)):
            raise ValueError("Duration must be a number")
        self._cpu_duration = value

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, value):
        if value and not isinstance(value, datetime.datetime):
            raise ValueError("Dates must be datetime.datetime objects")
        self._start_time = value

    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    def end_time(self, value):
        if value and not isinstance(value, datetime.datetime):
            raise ValueError("Dates must be datetime.datetime objects")
        self._end_time = value

    @property
    def map(self):
        d = {'VMUUID': self.uuid,
             'SiteName': self.site,
             'MachineName': self.name,
             'LocalUserId': self.user_id,
             'LocalGroupId': self.group_id,
             'FQAN': self.fqan,
             'Status': self.status,
             'StartTime': self.start_time and int(
                 self.start_time.strftime("%s")
             ),
             'EndTime': self.end_time and int(self.end_time.strftime("%s")),
             'SuspendDuration': self.suspend_duration,
             'WallDuration': self.wall_duration,
             'CpuDuration': self.cpu_duration,
             'CpuCount': self.cpu_count,
             'NetworkType': self.network_type,
             'NetworkInbound': self.network_in,
             'NetworkOutbound': self.network_out,
             'Memory': self.memory,
             'Disk': self.disk,
             'StorageRecordId': self.storage_record_id,
             'ImageId': self.image_id,
             'CloudType': self.cloud_type,
             'GlobalUserName': self.user_dn,
             'PublicIPCount': self.public_ip_count,
             'Benchmark': self.benchmark_value,
             'BenchmarkType': self.benchmark_type,
             'CloudComputeService': self.compute_service,
             }
        return d

    def as_json(self, version=None):
        return json.dumps(self.as_dict(version=version))



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from oslo_config import cfg
from oslo_log import log
import six

from caso import exception
from caso import loadables

CONF = cfg.CONF

LOG = log.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class BaseMessenger(object):
    @abc.abstractmethod
    def push(self, records):
        """Push the records."""


class Manager(loadables.BaseLoader):
    def __init__(self):
        super(Manager, self).__init__(BaseMessenger)
        self.messengers = None

    def _load_messengers(self):
        self.messengers = [i()
                           for i in self.get_matching_classes(CONF.messengers)]

    def push_to_all(self, records):
        if self.messengers is None:
            self._load_messengers()

        for m in self.messengers:
            try:
                m.push(records)
            except exception.RecordVersionNotFound:
                # Oops, a messenger is using a weird version, stop working
                LOG.error("Messenger '%s' is using an unknown "
                          "record version" % m.__class__.__name__)
                raise
            except Exception as e:
                # Capture exception so that we can continue working
                LOG.error(e)


def all_managers():
    return Manager().get_all_classes()



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import warnings

import dirq.QueueSimple
from oslo_config import cfg
from oslo_log import log
import six

import caso.messenger
from caso import utils

LOG = log.getLogger(__name__)

opts = [
    cfg.StrOpt('output_path',
               default='/var/spool/apel/outgoing/openstack',
               help="Directory to put the generated SSM records."),
    cfg.IntOpt('max_size',
               default=100,
               help="Maximum number of records to send per message"),
]

CONF = cfg.CONF

CONF.register_opts(opts, group="ssm")


__all__ = ["SsmMessager", "SSMMessengerV02", "SSMMessengerV04"]


@six.add_metaclass(abc.ABCMeta)
class _SSMBaseMessenger(caso.messenger.BaseMessenger):
    version = None
    separator = "%%"

    def __init__(self):
        # FIXME(aloga): try except here
        utils.makedirs(CONF.ssm.output_path)
        self.header = "APEL-cloud-message: v%s" % self.version

    def push(self, records):
        if not records:
            return

        entries = []
        for _, record in six.iteritems(records):
            aux = ""
            for k, v in six.iteritems(record.as_dict(version=self.version)):
                if v is not None:
                    aux += "%s: %s\n" % (k, v)
            entries.append(aux)

        # FIXME(aloga): try except here
        queue = dirq.QueueSimple.QueueSimple(CONF.ssm.output_path)

        # Divide message into smaller chunks as per GGUS #143436
        # https://ggus.eu/index.php?mode=ticket_info&ticket_id=143436
        for i in range(0, len(entries), CONF.ssm.max_size):
            message = "%s\n" % self.header

            sep = "%s\n" % self.separator
            message += "%s" % sep.join(entries[i:i + CONF.ssm.max_size])

            queue.add(message)


class SSMMessengerV02(_SSMBaseMessenger):
    version = "0.2"


class SSMMessengerV04(_SSMBaseMessenger):
    version = "0.4"


class SsmMessager(SSMMessengerV02):
    def __init__(self):
        msg = ("Using deprecated caso.messenger.ssm.SsmMessager, "
               "please use caso.messenger.ssm.SSMMessengerV02 if you "
               "wish to continue usinf the 0.2 version of the record, "
               "or refer to the cASO documentation.")
        warnings.warn(msg, DeprecationWarning)
        super(SsmMessager, self).__init__()
